/*
 * This is a module which is used for logging packets.
 */
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/spinlock.h>
#include <net/icmp.h>
#include <net/udp.h>
#include <net/tcp.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/netfilter_ipv4/ip_tables.h>

struct in_device;
#include <net/route.h>
#include <linux/proc_fs.h>
#include <linux/netfilter_ipv4/ipt_sLog.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#define NMACSIX(addr) \
	((unsigned char *)&addr)[0], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[3], \
	((unsigned char *)&addr)[4], \
	((unsigned char *)&addr)[5]
				
struct nat_devices_def
{
	unsigned long lanIP;
	unsigned long tick;
	unsigned char mac[ETH_ALEN];
	void*  	 next;
};
typedef struct nat_devices_def NAT_DEVICE_DEF;
static NAT_DEVICE_DEF* g_nat_devices;

static kmem_cache_t *ipt_sLog_cachep;
static unsigned long tickLimit;
        
/* Use lock to serialize, so access to shared memory don't overlap */
static spinlock_t slog_lock = SPIN_LOCK_UNLOCKED;


// Update the active user list
static void update_sLog(const struct ipt_sLog_info *info,
	struct iphdr *iph, unsigned long tick, unsigned char* mac)
{
	unsigned long count = 0;
	unsigned long max_num   = IPT_SLOG_MAX_ENTRIES;                  

	NAT_DEVICE_DEF* p = g_nat_devices;
	NAT_DEVICE_DEF* previous=0;

	unsigned long lanIP = iph->saddr;
	DEBUGP("SRC=%u.%u.%u.%u DST=%u.%u.%u.%u ",
	      NIPQUAD(iph->saddr), NIPQUAD(iph->daddr));

	tickLimit = IPT_SLOG_TIMEOUT*HZ;
	if( info )
	{
		tickLimit = info->timeout*HZ;
		max_num = info->max_num;
	}	

	// The match is performed by the rule from userspace
	//   This target is only applied to FORWARD
	//     state = ESTABLISHED?
	// use: iptables -A FORWARD -i br0 -j sLog    

	// list is empty
	if( p==0 )
	{
		// create one and insert
		p = (NAT_DEVICE_DEF*)kmem_cache_alloc(ipt_sLog_cachep, GFP_ATOMIC);
		memset(p, 0, sizeof(*p));
		p->lanIP = lanIP;
		p->tick  = tick;
		p->next  = 0;
		g_nat_devices = p;

		return;
        };

	// searching
	while( p )
	{
		// found
		if( p->lanIP==lanIP )
		{
			// just update
			if( tick>p->tick )
			{
				p->tick = tick;
			        memcpy(p->mac, mac, 6);
                        };

                        break;
                }
                // not match, exceed time limit
                else if( (tick - p->tick) > tickLimit )
                {
                        // idle for a long time:  delete this item
                        NAT_DEVICE_DEF* toDel = p;
                        if( previous==0 )                       // first item
                        {
                            // go to next
                            p = p->next;
                            // previous keep same value
                            previous = 0;
                            // delete first item must update g_nat_devices !!!
                            g_nat_devices = p;
                        }
                        else if( p->next==0 )                   // last item
                        {
                            // next will be 0
                            p = 0;
                            // previous keep value, but set its next to 0
                            previous->next = 0;
                        }
                        else                                    // inter item
                        {
                            // previous keep value, but set its next to p->next
                            previous->next = p->next;
                            // go to next
                            p = p->next;
                        };
			kmem_cache_free(ipt_sLog_cachep, toDel);
                }
                // not match, within time limit

                else
                {
                        previous = p;
                        // go to next item
                        p  = p->next;
                };

		count++;
        };

        // Reach the end, not found
        if( p==0 && count<max_num )
        {
                // add a new one to the tail
                p = (NAT_DEVICE_DEF*)kmem_cache_alloc(ipt_sLog_cachep, GFP_ATOMIC);
		memset(p, 0, sizeof(*p));
                p->lanIP = lanIP;
                p->tick  = tick;
		memcpy(p->mac, mac, 6);
                p->next  = 0;

                if( previous==0 )
                {
                        g_nat_devices = p;
                }
                else
                {
                        // chain it into the list
                        previous->next  = p;
                };
        };
}

static void destroy_sLog_table(void)
{
	NAT_DEVICE_DEF* p = g_nat_devices;
	while( p )
	{
		NAT_DEVICE_DEF* toDel = p;		
		p = p->next;
		kmem_cache_free(ipt_sLog_cachep, toDel);	
	}	
}

static int sLog_gethostname(unsigned long ip, char* name)
{
	/* query some name service, better in userspace */
	name[29] = '\0';
	return 0;
}	

int sLog_read_procmem(char *buf, char **start, off_t offset,
		                   int count, int *eof, void *data)
{
	int limit = count - 100; 
	int len = 0;
	unsigned long tick;
	
	// if (down_interruptible(sem))
	//	return -ERESTARTSYS;
	NAT_DEVICE_DEF* p;
       
	spin_lock_bh(&slog_lock);
	p = g_nat_devices;
	
	// len += sprintf(buf+len, "\nip address\tmac\tjiffies\tname\n");
	tick = jiffies;
	while( p && len<=limit )
	{
		if( (tick-p->tick) < tickLimit )
		{
			char name[30] = {0};	
			sLog_gethostname(p->lanIP, name);
			len += sprintf(buf+len, "%u.%u.%u.%u\t%02x:%02x:%02x:%02x:%02x:%02x\t%li\t%s\n",
				NIPQUAD(p->lanIP), NMACSIX(p->mac), p->tick, name);
		}	
		p = p->next;
	}
	
	spin_unlock_bh(&slog_lock);	
	// up(sem);	
	*eof = 1;
	return len;
}


static void sLog_create_proc()
{
	create_proc_read_entry("net/ipt_sLog", 0 /* default mode */,
		NULL /* parent dir */, sLog_read_procmem,
		NULL /* client data */);
}

static void sLog_remove_proc()
{
	/* no problem if it was not registered */
	remove_proc_entry("net/ipt_sLog", NULL /* parent dir */);
}

static unsigned int
ipt_sLog_target(struct sk_buff **pskb,
	       unsigned int hooknum,
	       const struct net_device *in,
	       const struct net_device *out,
	       const void *targinfo,
	       void *userinfo)
{
	struct iphdr *iph = (*pskb)->nh.iph;
	const struct ipt_sLog_info *loginfo = targinfo;
	unsigned char mac[ETH_ALEN] = {0};

	spin_lock_bh(&slog_lock);
	DEBUGP("IN=%s OUT=%s ", in ? in->name : "", out ? out->name : "");
	if (in && out) 
	{
	/*	if ((*pskb)->dev && (*pskb)->dev->hard_header_len && (*pskb)->mac.raw != (void*)iph) {
			unsigned char *p = (*pskb)->mac.raw;
			memcpy(mac, p, 6);
		}  */
		/* in: br0 use h_source;	out: br0 use h_destination	*/
        	memcpy(mac,(*pskb)->mac.ethernet->h_source, ETH_ALEN);
		update_sLog(loginfo, iph, jiffies, mac);
	}
	DEBUGP("\n");

	spin_unlock_bh(&slog_lock);

	return IPT_CONTINUE;
}

static int ipt_sLog_checkentry(const char *tablename,
			      const struct ipt_entry *e,
			      void *targinfo,
			      unsigned int targinfosize,
			      unsigned int hook_mask)
{
	const struct ipt_sLog_info *loginfo = targinfo;

	if ( targinfosize == IPT_ALIGN(0) )
		return 1;
	
	if ( targinfosize != IPT_ALIGN(sizeof(struct ipt_sLog_info))) {
		printk("sLog: targinfosize %u != %u\n",
		       targinfosize, IPT_ALIGN(sizeof(struct ipt_sLog_info)));
		return 0;
	}

        if (loginfo->max_num > IPT_SLOG_MAX_ENTRIES) {
	        printk("sLog: max_num %u > %u\n", loginfo->max_num, IPT_SLOG_MAX_ENTRIES);
		return 0;
	}

	return 1;
}

static struct ipt_target ipt_sLog_reg
= { { NULL, NULL }, "sLog", ipt_sLog_target, ipt_sLog_checkentry, NULL, 
    THIS_MODULE };

static int __init init(void)
{
	if (ipt_register_target(&ipt_sLog_reg))
		return -EINVAL;
	
        ipt_sLog_cachep = kmem_cache_create("ipt_sLog",
		sizeof(NAT_DEVICE_DEF), 0, SLAB_HWCACHE_ALIGN, NULL, NULL);
	if (!ipt_sLog_cachep) {
		printk(KERN_ERR "Unable to create ipt_sLog slab cache\n");
		return -ENOMEM;
	}
	
	sLog_create_proc();

	return 0;
}

static void __exit fini(void)
{
	sLog_remove_proc();

	// destroy the hash table
	destroy_sLog_table();

	kmem_cache_destroy(ipt_sLog_cachep);

	ipt_unregister_target(&ipt_sLog_reg);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
